@@ -5,7 +5,7 @@
  * This file contains the implementation of KBiffMonitor and
  * associated classes.
  *
- * $Id: kbiffmonitor.cpp,v 1.17 2003/12/19 21:32:41 gran Exp $
+ * $Id: kbiffmonitor.cpp,v 1.19 2005/06/02 03:31:10 gran Exp $
  */
 #include "kbiffmonitor.h"
 
@@ -43,6 +43,10 @@
 #include <qdatetime.h>
 #include <ksimpleconfig.h>
 
+// Needed for CRAM-MD5 and APOP
+#include <kmdcodec.h>
+#include "kbiffcrypt.h"
+
 #define MAXSTR (1024)
 
 #define MAIL_STATE_FILE "kbiffstate"
@@ -261,9 +265,9 @@ void KBiffMonitor::setMailbox(KBiffURL& url)
         mailbox  = url.path().right(url.path().length() - 1); 
         port     = (url.port() > 0) ? url.port() : 143;
 
-        preauth = url.query() == "?preauth" ? true : false;
-        keepalive = url.ref() == "keepalive" ? true : false;
-        bool async = url.ref() == "async" ? true : false;
+        preauth = url.searchPar("preauth") == "yes";
+        keepalive = url.searchPar("keepalive") == "yes";
+        bool async = url.searchPar("async") == "yes";  
         imap->setAsync(async);
 #ifdef USE_SSL
         imap->setSSL(false);
@@ -286,9 +290,9 @@ void KBiffMonitor::setMailbox(KBiffURL& url)
         mailbox  = url.path().right(url.path().length() - 1); 
         port     = (url.port() > 0) ? url.port() : 993;
 
-        preauth = url.query() == "?preauth" ? true : false;
-        keepalive = url.ref() == "keepalive" ? true : false;
-        bool async = url.ref() == "async" ? true : false;
+        preauth = url.searchPar("preauth") == "yes";
+        keepalive = url.searchPar("keepalive") == "yes";
+        bool async = url.searchPar("async") == "yes";
         imap->setAsync(async);
 	imap->setSSL(true);
         simpleURL = "imap4s://" + server + "/" + mailbox;
@@ -308,8 +312,8 @@ void KBiffMonitor::setMailbox(KBiffURL& url)
         mailbox  = url.user();
         port     = (url.port() > 0) ? url.port() : 110;
 
-        keepalive = url.ref() == "keepalive" ? true : false;
-        bool async = url.ref() == "async" ? true : false;
+        keepalive = url.searchPar("keepalive") == "yes";
+        bool async = url.searchPar("async") == "yes";
         pop->setAsync(async);
 #ifdef USE_SSL
         pop->setSSL(false);
@@ -332,8 +336,8 @@ void KBiffMonitor::setMailbox(KBiffURL& url)
         mailbox  = url.user();
         port     = (url.port() > 0) ? url.port() : 995;
 
-        keepalive = url.ref() == "keepalive" ? true : false;
-        bool async = url.ref() == "async" ? true : false;
+        keepalive = url.searchPar("keepalive") == "yes";
+        bool async = url.searchPar("async") == "yes";
         pop->setAsync(async);
 	pop->setSSL(true);
 
@@ -395,8 +399,8 @@ void KBiffMonitor::setMailbox(KBiffURL& url)
         mailbox  = url.path().right(url.path().length() - 1); 
         port     = (url.port() > 0) ? url.port() : 119;
 
-        keepalive = url.ref() == "keepalive" ? true : false;
-        bool async = url.ref() == "async" ? true : false;
+        keepalive = url.searchPar("keepalive") == "yes";
+        bool async = url.searchPar("async") == "yes";
         nntp->setAsync(async);
 #ifdef USE_SSL
         nntp->setSSL(false);
@@ -550,16 +554,14 @@ void KBiffMonitor::checkPop()
             return;
         }
 
-        command = "USER " + user + "\r\n";
-        if (pop->command(command) == false)
-        {
-            pop->close();
-            invalidLogin();
-            return;
-        }
-
-        command = "PASS " + password + "\r\n";
-        if (pop->command(command) == false)
+        // find out if APOP is supported
+        pop->parseBanner();
+        
+        // find other possibly useful capabilities
+        // we don't care if this fails
+        pop->command("CAPA\r\n");
+        
+        if (pop->authenticate(user, password) == false )
         {
             pop->close();
             invalidLogin();
@@ -624,20 +626,9 @@ void KBiffMonitor::checkImap()
         }
 
         do_login = true;
-    }
-    
-    // imap allows spaces in usernames... we need to take care of that
-    user = imap->mungeUserPass(user);
 
-    // also asterisks (*) in passwords. maybe it's a good idea
-    // to _always_ munge the user and the password.
-    password = imap->mungeUserPass(password);
-
-    // if we are preauthorized OR we want to keep the session alive, then
-    // we don't login.  Otherwise, we do.
-    if ((preauth == false) && (do_login == true))
-    {
-        command = QString().setNum(seq) + " LOGIN " + user + " " + password + "\r\n";
+        // check the server's capabilities (see RFC 3050, 6.1.1)
+        command = QString().setNum(seq) + " CAPABILITY\r\n";
         if (imap->command(command, seq) == false)
         {
             invalidLogin();
@@ -646,6 +637,17 @@ void KBiffMonitor::checkImap()
         seq++;
     }
 
+    // if we are preauthorized OR we want to keep the session alive, then
+    // we don't login.  Otherwise, we do.
+    if ((preauth == false) && (do_login == true))
+    {
+        if (imap->authenticate(&seq, user, password) == false)
+        {
+            invalidLogin();
+            return;
+        }
+    }
+
     // reset the numbers from the last check
     imap->resetNumbers();
     
@@ -1560,6 +1562,7 @@ bool KBiffSocket::connectSocket(const QString& host, unsigned short int port)
     }
 
     // everything is swell
+    banner = line;  // save the banner for use by subclasses
     return true;
 }
 
@@ -1711,6 +1714,12 @@ QString KBiffSocket::readLine()
 ///////////////////////////////////////////////////////////////////////////
 // KBiffImap
 ///////////////////////////////////////////////////////////////////////////
+KBiffImap::KBiffImap()
+{
+    /* Assume that the IMAP server does no fancy authentication */
+    auth_cram_md5 = false;
+}
+
 KBiffImap::~KBiffImap()
 {
     close();
@@ -1722,6 +1731,7 @@ bool KBiffImap::command(const QString& line, unsigned int seq)
     bool unseen = false;
     QString messagesListString;
     QStringList messagesList;
+    bool tried_cram_md5; // are we trying CRAM-MD5 ?
 
     if (writeLine(line) <= 0)
     {
@@ -1736,6 +1746,12 @@ bool KBiffImap::command(const QString& line, unsigned int seq)
     
     // must be case insensitive
     QRegExp status("\\* STATUS", FALSE);
+    QRegExp capability("\\* CAPABILITY", FALSE);
+    QRegExp cram_md5("AUTHENTICATE CRAM-MD5", FALSE);
+
+    // are we trying CRAM-MD5 ?
+    tried_cram_md5 = cram_md5.search(line)>=0;
+    cram_md5 = QRegExp("\\+ ([A-Za-z0-9+/=]+)");
     
     while (!(response = readLine()).isNull())
     {
@@ -1770,6 +1786,31 @@ bool KBiffImap::command(const QString& line, unsigned int seq)
                 messages = num.toInt();
             }
         }
+
+        /* The CAPABILITY response is documented in RFC 3050,
+         * sections 6.1.1 and 7.2.1
+         * An example:
+         * C: . CAPABILITY
+         * S: * CAPABILITY IMAP4rev1 IDLE AUTH=PLAIN AUTH=CRAM-MD5
+         * S: . OK CAPABILITY completed.
+         */
+        if (capability.search(response) >= 0) {
+            QRegExp cram_md5("AUTH=CRAM-MD5", FALSE);
+            if (cram_md5.search(response) >= 0) {
+                auth_cram_md5 = true;
+            }
+        }
+
+        /* AUTHENTICATE CRAM-MD5 response is documented in
+         * RFC 3050 6.2.2 and RFC 2195
+         */
+        if (tried_cram_md5 && cram_md5.search(response)>=0) {
+            chall_cram_md5 = KCodecs::base64Decode(cram_md5.cap(1).local8Bit());
+            if (chall_cram_md5.isNull())
+                break;
+
+            return true;
+        }
     }
 
     close();
@@ -1794,6 +1835,55 @@ void KBiffImap::resetNumbers()
     newMessages = 0;
 }
 
+bool KBiffImap::authenticate(int *pseq, const QString& user, const QString& pass)
+{
+    QString cmd, username, password;
+
+    // If CRAM-MD5 is available, use it.  It's the best we know.
+    // RFC 2195 defines the CRAM-MD5 authentication method
+    // also see RFC 3501 section 6.2.2 for the AUTHENTICATE command
+    if( auth_cram_md5 )
+    {
+        cmd = QString("%1 AUTHENTICATE CRAM-MD5\r\n").arg(*pseq);
+        if (command(cmd, *pseq) == false)
+        {
+            return false;
+        }
+
+        // calculate the real response to the challenge
+        QString response = user + " " + KBiffCrypt::hmac_md5(chall_cram_md5, pass);
+        response = KCodecs::base64Encode(response.latin1());
+
+        // send the response
+        if (command(response+"\r\n", *pseq) == false)
+        {
+            return false;
+        }
+
+        return true;
+    }
+    else // well, we tried, LOGIN is the best we can do
+    {
+        // imap allows spaces in usernames... we need to take care of that
+        username = mungeUserPass(user);
+
+        // also asterisks (*) in passwords. maybe it's a good idea
+        // to _always_ munge the user and the password.
+        password = mungeUserPass(pass);
+
+        cmd = QString().setNum(*pseq) + " LOGIN "
+            + username + " "
+            + password + "\r\n";
+        if (command(cmd, *pseq) == false)
+        {
+            return false;
+        }
+        (*pseq)++;
+    }
+
+    return true;
+}
+
 ///////////////////////////////////////////////////////////////////////////
 // KBiffPop
 ///////////////////////////////////////////////////////////////////////////
@@ -1853,6 +1943,34 @@ bool KBiffPop::command(const QString& line)
         if (!response.isNull())
             sscanf(response.ascii(), "+OK %d", &messages);
     }
+    else
+    // find out what the server is capable of
+    if (line == "CAPA\r\n")
+    {
+        QRegExp rx("\\bCRAM-MD5\\b");
+
+        auth_cram_md5 = false;  // assume no support
+
+        for (response = readLine();
+             !response.isNull() && response.left(1) != ".";
+             response = readLine())
+        {
+            if (response.left(4) == "SASL")
+                auth_cram_md5 = response.find(rx) != -1;
+        }
+    }
+    else
+    // look for the CRAM-MD5 challenge
+    if (line == "AUTH CRAM-MD5\r\n")
+    {
+        QRegExp challenge("\\+ ([A-Za-z0-9+/=]+)");
+        if (challenge.search(response) == -1 )
+        {
+            return false;
+        }
+
+        chall_cram_md5 = KCodecs::base64Decode(challenge.cap(1).local8Bit()); 
+    }
 
     return !response.isNull();
 }
@@ -1862,6 +1980,99 @@ KBiffUidlList KBiffPop::getUidlList() const
     return uidlList;
 }
 
+/*!
+   This method parses the initial response from the POP3 server.
+   The response is defined in RFC 1939 sections 4 and 7.
+
+    \fn KBiffPop::parse_banner(void)
+ */
+bool KBiffPop::parseBanner(void)
+{
+    // RFC 1939 section 3 says server MUST use uppercase
+    if( banner.left(3) != "+OK" ) {
+        auth_apop = false;
+        return false;
+    }
+
+    // Look for the banner part that indicates APOP support
+    QRegExp rx("(<[a-zA-Z0-9_+.-]+@[a-zA-Z0-9_+.-]+>)");
+    if( rx.search(banner) == -1 ) {
+        auth_apop = false;
+    } else {
+        chall_apop = rx.cap(1).latin1();
+        auth_apop  = true;
+    }
+
+    return true;
+}
+
+/*!
+    This method authenticates using the most secure
+    technique available.
+    \fn KBiffPop::authenticate(const QString& user, const QString& pass)
+ */
+bool KBiffPop::authenticate(const QString& user, const QString& pass)
+{
+    QString command;
+
+    // CRAM-MD5 authentication is the most secure we can handle
+    // the use of the AUTH command is documented in RFC 1734
+    if( auth_cram_md5 )
+    {
+        if (this->command("AUTH CRAM-MD5\r\n") == false)
+        {
+            return false;
+        }
+
+        // calculate the real response to the challenge
+        QString response = user + " " + KBiffCrypt::hmac_md5(chall_cram_md5, pass);
+        response = KCodecs::base64Encode(response.latin1());
+
+        // send the response
+        if (this->command(response+"\r\n") == false)
+        {
+            return false;
+        }
+        
+        return true;
+    }
+
+    // APOP is not as secure as CRAM-MD5 but it's still better
+    // than sending the password in the clear
+    if( auth_apop )
+    {
+        QCString digest;
+
+        KMD5 md5(chall_apop);
+        md5.update(pass);
+
+        digest = md5.hexDigest();
+        
+        command = QString("APOP %1 %2\r\n").arg(user, digest.data());
+        if (this->command(command) == false)
+        {
+            return false;
+        }
+        
+        return true;
+    }
+    
+    // lastly we'll try regular, plain-text authentication
+    
+    command = "USER " + user + "\r\n";
+    if (this->command(command) == false)
+    {
+        return false;
+    }
+
+    command = "PASS " + pass + "\r\n";
+    if (this->command(command) == false)
+    {
+        return false;
+    }
+    
+    return true;
+}
 ///////////////////////////////////////////////////////////////////////////
 // KBiffNntp
 ///////////////////////////////////////////////////////////////////////////
